/*->h.odef */


/***************************************************************************/

typedef struct oareabox
{
 int xlo;
 int ylo;
 int xhi;
 int yhi;
} oareabox;


/***************************************************************************/
                        /* structure of vm memory block */

typedef struct obuffer
{
  struct obuffer **  anchor;  /* this points at the anchor in the source   */
  int     len;                /* amount of stuff in buffer */
  int     maxlen;             /* maximum length of buffer  */
  int     previous;           /* offset of previous buffer */
  int     goffset;            /* offset of gap in buffer   */
  char    buff[4];            /* character buffer          */
} obuffer;

/***************************************************************************/



/*********** Universal file header structure  ********/
/*********** contains program specific things ********/

typedef struct
{
 char name[12];
 char  mark[8];
 char copy[64];

 char stuff[172];

} ouhead;

/********** Generic file header structure   *********/
/********** contains file specific things   *********/


typedef struct
{

 char stuff[256];


} oghead;

/***********  New page structure ********************/

typedef struct
{
 int size;
 int width;
 int height;
 int shape;
 int dble;
 int columns;
 int gutter;
 int topmargin;
 int botmargin;
 int leftmargin;
 int rightmargin;
 int headermargin;
 int footermargin;
} onewpage;


/************ Print structures ***********************/


typedef struct
{
 int copies;
 int from;
 int to;
 int all;
 int oddevenall;
 int backwards;
 int collate;
 int pause;
 int printpics;
 int draft;
 int merge;
 char froms[12];
 char tos[12];
} o1printstr;



typedef struct
{
 int flags;
 int xmul;
 int xdiv;
 int ymul;
 int ydiv;
 int cx;
 int cy;
} opset;


/************ Zoom structure ************************/

typedef struct
{
 int mul;
 int div;
} ozoomer;


/*********** Static paragraph structure  ************/

#define XLEADABS  0x1
#define XLEADPC   0x2
#define XLEADDL   0x4
#define XRIGHT    0x8
#define XLEFT     0x10
#define XCENTRE   0x20
#define XJUSTIFY  0x40
#define XWORDWRAP 0x80
#define XHYPH     0x100
#define XHYCAPS   0x200
#define XAUTOKERN 0x400


#define XMAXTABS  100



typedef struct opara
{
 int leading;           /* leading in appropriate units    */
 int flags;             /* flags as above                  */
 int first;             /* first line indent               */
 int left;              /* left indent                     */
 int right;             /* right indent                    */
 int before;            /* space before                    */
 int after;             /* space after                     */
 int minwlen;           /* min word length for hypehnation */
 int ntab;              /* number of tabs                  */

 struct
        {
         int type;
         int posn;
        } tabs[XMAXTABS];

} oparastr;



typedef struct oparastore
{
 int leading;           /* leading in appropriate units */
 int flags;             /* flags                        */
 int first;             /* first line indent            */
 int left;              /* left indent                  */
 int right;             /* right indent                 */
 int before;            /* space before                 */
 int after;             /* space after                  */
 int minwlen;           /* min word len for hyphenation */
 int ntab;              /* number of tabs               */
 int toff;              /* offset of tabs for this para */
 int count;             /* times para has been added    */
 int para;              /* number of para               */
} oparastore;



typedef struct ofileparastore
{
 int        noparas;    /* number of paras stored      */
 int        notabs;     /* number of tabs stored       */
 obuffer *   paras;      /* anchor for storage of paras */
 obuffer *   tabs;       /* anchor for storage of tabs  */
} ofileparastore;



/************************************************************************/
/* structure to handle font styles  */

typedef struct oparastyle  /* a style vector */
{
 int parano;   /* our number for the paragraph style */
 int   font;   /* our number for the font name       */
 int   xpts;   /* xsize in kpts                      */
 int   ypts;   /* ysize in kpts                      */
 int  track;   /* absolute tracking in kpts          */
 int  style;   /* style bits                         */
 int   fcol;   /* foreground colour                  */
 int   lead;   /* absolute leading for the font      */
 int height;   /* absolute height for the font       */
} oparastyle;



typedef struct ouserparastyle    /* a user style */
{
 char       name[12];
 int        key;
 oparastyle  ps;
} ouserparastyle;



typedef struct ofileupstyles
{
 int      nostyles;
 obuffer * anchor;
} ofileupstyles;



typedef struct ostyle
{
 int        number;     /* number for this fstyle                   */
 int        count;      /* times this style is used                 */
 int        upstyleno;  /* number of user style                     */
 int        bits;       /* bits to show differences from upstyle no */
 oparastyle  pstyle;     /* absolute style vector                    */
} ostyle;



typedef struct ofilesstyles
{
 int          no;       /* number of style entries */
 obuffer * anchor;       /* buffer to hold them     */ 
} ofilestyles;




/************* Vline structure **********************/

typedef struct
{
 int          xxx;
 int          xxy;
 int          xxw;
 int          xxh;
 int          xmw;   /* maximum formated width */
} ovline;

/************* Box structure ************************/


typedef struct xstr
{
 struct xstr * next;
 struct xstr * prev;
 char          type;                                           /* X type  */
 char            f1;                                           /* flags   */
 char            f2;
 char            f3;  
 int        x,y,w,h;                                           /* 7 words */

   union  {

         struct {                        /* 1 word */
                 struct xstr * box;
                } vbox;


         struct {                         /* 8 words */
                 struct xstr * textbox;
                 obuffer *     txanc;     /* anchor for text buffer         */
                 obuffer *     vlanc;     /* anchor for line buffer         */
                 int           foffset;   /* highest flowed point           */
                 int           doffset;   /* point that must be reflowed to */
                 int           hoffset;   /* highest unflowed point         */
                 int           qoffset;   /* quick reflow                   */
                 struct xstr * link;      /* temporary link                 */

                } source;
                  

         struct {                        /* 7 words */
                 struct xstr * picbox;
                 obuffer *     anchor;
                 int           type;
                 int           x0;
                 int           y0;
                 int           x1;
                 int           y1;
                } picture;
      

         struct {                        /* 14 words           */
                 struct xstr * parent;
                 struct xstr * child;
                 struct xstr * frame;    /* frame pointer      */
                 struct xstr * fnext;    /* flow next          */
                 struct xstr * fprev;    /* flow prev          */
                 int           colour;   /* text colour        */
                 int           trans;    /* transparent ?      */
                 int           i;        /* inset              */
                 int           o;        /* outset             */

                 struct xstr * vbox;     /* first vibox        */
                 int           bytes;    /* bytes in  box      */
                 int           lines;    /* lines in  box      */
                 int           influx;   /* inflow to box      */
                 int           linflux;  /* inflow in lines    */
                 struct xstr * oldvbox;
                } text;
                       

         struct {                        /* 17 words                      */
                 struct xstr * parent;
                 struct xstr * child;
                 struct xstr * frame;    /* frame pointer                 */
                 struct xstr * fnext;    /* next pic box with this source */
                 struct xstr * fprev;    /* prev pic box with this source */
                 int           colour;   /* picture colour                */
                 int           trans;    /* transparent ?                 */
                 int           i;        /* inset                         */
                 int           o;        /* outset                        */

                 int           file;     /* file handle for this picbox   */
                 int           xmul;     /* xscale                        */
                 int           xdiv;     
                 int           ymul;     /* yscale                        */
                 int           ydiv;
                 int           xshift;   /* x shift                       */
                 int           yshift;   /* y shift                       */
                 int           angle;    /* angle                         */
                } pict;


         struct
                {                         /* 14 words */
                 struct xstr * parent;
                 int           colour;
                 int           style;
                 int           endcaps;
                 int           ox;
                 int           oy;
                 int           ex;
                 int           ey;
                 int           thickness;

                 int           bx;
                 int           by;
                 int           bh;
                 int           bw;
                 int           hy;

                } line;


         struct
                {                        /* 7 words */
                 struct xstr * parent;
                 int           colour;   /* frame colour       */
                 int           size;     /* frame size         */
                 int           spcol;    /* use sprite colours */
                 char          name[12]; /* name of frame      */
                } frame;


         struct {
                 struct xstr * parent;    /* so it looks like a text box ! */
                 struct xstr * child;
                 struct xstr * gux;
                 struct xstr * guy;
                 int           number;    /* running  page  number */
                 int           guxa[2];
                 int           guya[2];
                } page;


         struct {                      /* 13 words */
                 struct xstr * page;   /* first page in section          */
                 struct xstr * source; /* first source in section        */
                 int np;               /* number of pages                */
                 short int secno;      /* section number                 */
                 char  format;         /* page number format             */
                 char  start;          /* 1 == start numbering from here */
                 int number;           /* starting number                */

                 char prs[32];         /* pointer to printer string */
                } section;


         struct {                       /* 4 words */
                 struct xstr * parent;  /* points back to page       */
                 struct xstr * child;   /* points to any child box's */
                 struct xstr * frame;   /* frame                     */
                 struct xstr * sub;     /* points to sub text box's  */
                } shell;

         } u;

} xstr;




/* f1 flags                   */

#define XMODHSECT 0x1
#define XMODFSECT 0x2
#define XFULLBOX  0x4
#define XCLEAROBJ 0x8



typedef struct ovec              /* position vector  */
{
 xstr *     section;            /* section          */
 xstr *     source;             /* source           */
 int        off;                /* offset in source */

 xstr *     page;               /* page             */
 xstr *     box;                /* box              */

 int        vloff;              /* vline offset in vl buff */
 int        loff;               /* line offset in source   */
 int        vlboff;             /* box  offset in vl buff  */
 int        boff;               /* box  offset in source   */

 int        fstyle;             /* fstyle at start of line */

 int        xps;                /* coords of page  */
 int        yps;
 int        x;                  /* coords of point */
 int        y;

 int        ld;                 /* fmaxld for line */

 int        vdi;                /* vdi for point   */
} ovec;



#define XPAGE    0 
#define XTEXT    1
#define XPICT    2 
#define XLINE    3 
#define XMAIN    4 
#define XHEADER  5
#define XFOOTER  6 
#define XVBOX    7 
#define XSECT    8

#define XACTOUT  10 
#define XACTRHM  11 
#define XACTRHB  12 
#define XACTBTM  13 
#define XGUIDEX  14 
#define XGUIDEY  15 

#define XACTRHT  17
#define XACTTPM  18
#define XACTLHT  19
#define XACTLHM  20
#define XACTLHB  21
#define XSOURCE  22 
#define XFRAME   23
#define XSHELL   24
#define XPICTURE 25

#define XACTLO   26
#define XACTLE   27

#define XFREE32  28
#define XFREE64  XFREE32+1
#define XFREE96  XFREE64+1
#define XFREE128 XFREE96+1




#define XACBSUB  10


#define XACTMINW 6000
#define XACTMINH 6000


/* size of box full marker */

#define XFBOXW   12000
#define XFBOXH   12000


#define XULPOSN  -32            /* parameters for underline  */
#define XULHEIGHT  8

                               /* codes that appear in docs */
#define XXEOT      0

#define XXEOBOX    2
#define XXEOPAGE   3
#define XXTABCH    9
#define XXNL      10
#define XXCR      13
#define XXTABSEQ  18
#define XXMACSEQ  19
#define XXSNO     20
#define XXDCHYPH  21
#define XXSPACE   32
#define XXNBSPACE 160
#define XXNBHYPH  173
#define XXHYPHEN  45
#define XXPAGENO  30


#define XXTSLEN    4
#define XXMSLEN    4
#define XXSNLEN    4
                              /* font manager codes */



/* flags for font style */

#define XREVERSE 0x1
#define XUNDER   0x2
#define XWUNDER  0x4
#define XSUBS    0x8
#define XSUPS    0x10
#define XTITLE   0x20
#define XALLCAPS 0x40
#define XSMLCAPS 0x80
#define XITALICS 0x100
#define XBOLD    0x200



/* vline flags          */

#define XISJUST  0x03000000     /* line is justified                */
#define XISCENT  0x02000000     /* line is centered                 */
#define XISRIGHT 0x01000000     /* line is right justified          */
#define XISLEFT  0x00000000     /* line is left  justified          */

#define XCLEAR   0x80000000     /* line does not exist on screen    */
#define XHASPARA 0x04000000     /* line starts with a para code seq */
#define XHASFONT 0x08000000     /* this line has a font code in it  */

#define XISSPACE 0x00000000     /* line ends with a space           */
#define XISBREAK 0x10000000     /* line ends with a break           */
#define XISNL    0x20000000     /* line ends with a NL              */
#define XISTEXT  0x30000000     /* line ends with ordinary char     */




#define XSETVISIBLE(line)  ((line)->xmw=((line)->xmw & 0x7FFFFFFF))
#define XSETCLEAR(line)    ((line)->xmw=((line)->xmw | 0x80000000))
#define XLCLEAR(line)      (((line)->xmw & 0x80000000))

#define XSETEOL(line,code) ((line)->xmw=((line)->xmw & 0x8FFFFFFF)|(code))
#define XEOLCHAR(line)     (((line)->xmw) & 0x70000000)

#define XSETHYPH(line)     ((line)->xmw=((line)->xmw | 0x04000000))
#define XLHYPH(line)       (((line)->xmw & 0x04000000)) 

#define XSETFONT(line)     ((line)->xmw=((line)->xmw | 0x08000000))
#define XLFONT(line)       (((line)->xmw & 0x08000000))

#define XSETJ(line,j)      ((line)->xmw=((line)->xmw & 0xFCFFFFFF)|(j))
#define XLJ(line)          (((line)->xmw & 0x03000000))

#define XSETAUTOKERN(line) ((line)->xmw=((line)->xmw | 0x800000))
#define XLAUTOKERN(line)   (((line)->xmw) & 0x800000)



#define XSETX(line,x)      ((line)->xxx=((line)->xxx & 0xFF800000)|(x))    
#define XLX(line)          (((line)->xxx) & 0x7FFFFF)

#define XSETY(line,y)      ((line)->xxy=((line)->xxy & 0xFF800000)|(y))
#define XLY(line)          (((line)->xxy) & 0x7FFFFF)

#define XSETH(line,h)      ((line)->xxh=((line)->xxh & 0xFF800000)|(h))
#define XLH(line)          (((line)->xxh) & 0x7FFFFF)

#define XSETW(line,w)      ((line)->xxw=((line)->xxw & 0xFF800000)|(w))
#define XLW(line)          (((line)->xxw) & 0x7FFFFF)

#define XSETMW(line,m)     ((line)->xmw=((line)->xmw & 0xFF800000)|(m))
#define XLMW(line)         (((line)->xmw) & 0x7FFFFF)

#define XSETFS(line,f) \
   ((line)->xxx=(((line)->xxx & 0x00FFFFFF) | (((f) & 0xFF)<<24)),\
    (line)->xxy=(((line)->xxy & 0x00FFFFFF) | (((f) & 0xFF00)<<16)),\
    (line)->xxw=(((line)->xxw & 0x3FFFFFFF) | (((f) & 0x30000)<<14)))

#define XLFS(line)         ((((line)->xxx & 0xFF000000) >> 24) | \
                           (((line)->xxy & 0xFF000000) >> 16) | \
                           (((line)->xxw & 0xC0000000) >> 14))

#define XSETLEN(line,l) \
   ((line)->xxh=(((line)->xxh & 0x00FFFFFF) | (((l) & 0xFF)<<24)),\
    (line)->xxw=(((line)->xxw & 0xC0FFFFFF) | (((l) & 0x3F00)<<16)))

#define XLLEN(line)       ((int)((((line)->xxh & 0xFF000000) >> 24) | \
                               (((line)->xxw & 0x3F000000) >> 16) ))


/* buffer macros */

#define  XSIZEBUFF(bp) ((bp)->len-sizeof(buffer)+4)
#define  XSRCLEN(s)    SIZEBUFF((s)->u.source.txanc)
#define  XPICLEN(p)    SIZEBUFF((p)->u.picture.anchor)
#define  XVLLEN(s)     SIZEBUFF((s)->u.source.vlanc)
